import os
import sys
import shutil
import copy
import glob
import re
import csv
import numpy as np
from numpy import sin, cos, tan, arccos, arcsin, arctan, sqrt, exp, log, pi
from pprint import pprint

from tklib.tkobject import tkObject
from tklib.tkutils import IsDir, IsFile, SplitFilePath, pconv, pint, pfloat, safe_getelement
from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg
from tklib.tkutils import SplitFilePath, GetList
from tklib.tkre    import DelQuote, DelSpace
import tklib.tkre as tkre
from tklib.tkfile import tkFile
from tklib.tkcrystal.tkcrystalobject import tkCrystalObject
from tklib.tkcrystal.tkspacegroup import tkSpaceGroup
from tklib.tkprogvars import ProgramDir, DBDir, AtomDBDir, VASPDir, VASPPerlDir, VASPPythonDir
from tklib.tkfile import tkFile
from tklib.tkinifile import tkIniFile
from tklib.tksci.tksci import Reduce01, Round
from tklib.tksci.tksci import a0, NA, torad, todeg, Round, Factorize, IsFactor
from tklib.tksci.tksci import degcos, degsin, degtan, acos,asin, atan, degacos, degasin, degatan
from tklib.tksci.tkintegration import rieman_array_func
from tklib.tksci.tkmatrix import make_matrix1, make_matrix2, make_matrix3
from tklib.tkcrystal.tkcrystal import tkCrystal


conv_dic = {
          "alpha": 1
        , "beta": 1
        , "gamma": 1
        , "delta": 1
        , "epsilon": 1
        , "zeta": 1
        , "eta": 1
        , "theta": 1
        , "iota": 1
        , "kappa": 1
        , "lambda": 1
        , "mu": 1
        , "nu": 1
        , "xi": 1
        , "pi": 1
        , "rho": 1
        , "sigma": 1
        , "tau": 1
        , "upsilon": 1
        , "phi": 1
        , "chi": 1
        , "psi": 1
        , "omega": 1
        }


class tkVASP(tkCrystalObject):
    def __init__(self, **args):
        self.crystal     = None
        self.incarinf    = None
        self.kpointsinf  = None
        self.outcarinf   = None
        self.eigenvalinf = None

        self.update(**args)

    def __del__(self):
        pass

    def __str__(self):
        return self.ClassPath()

    def __del__(self):
        pass


    def getdir(self, path):
        if IsDir(path):
            return path

        dirname, basename, filebody, ext = SplitFilePath(path)
        return dirname

    def get_INCAR(self, CAR_dir):
        return os.path.join(CAR_dir, 'INCAR')

    def get_POSCAR(self, CAR_dir):
        return os.path.join(CAR_dir, 'POSCAR')

    def get_POTCAR(self, CAR_dir):
        return os.path.join(CAR_dir, 'POTCAR')

    def get_KPOINTS(self, CAR_dir):
        return os.path.join(CAR_dir, 'KPOINTS')

    def get_CONTCAR(self, CAR_dir):
        return os.path.join(CAR_dir, 'CONTCAR')

    def get_OUTCAR(self, CAR_dir):
        return os.path.join(CAR_dir, 'OUTCAR')

    def get_DOSCAR(self, CAR_dir):
        return os.path.join(CAR_dir, 'DOSCAR')

    def get_EIGENVAL(self, CAR_dir):
        return os.path.join(CAR_dir, 'EIGENVAL')

    def get_VASPPath(self, CAR_dir, filename):
        return os.path.join(CAR_dir, filename)

    def read_potcar_inf(self, potcarpath):
        inf = {}

        fp = tkFile(potcarpath, 'r')
        if not fp or fp.fp is None:
            return None

        count = 0
        while 1:
            line = fp.ReadLine()
            if not line:
                break

#  PAW_PBE O 08Apr2002                    
            line = line.strip()
            ret = tkre.Search(r'\S+\s+([A-Z][a-z]?)', line)
            if ret:
                inf["PP{}".format(count)] = line
                inf["Atom{}".format(count)] = ret[1]
                line = fp.ReadLine()
                inf["Ne{}".format(count)] = pfloat(line)
                fp.SkipTo("End of Dataset")
                count += 1

        inf["nPPTypes"] = count
        self.potcarinf = inf
         
        return inf

    def read_poscar(self, poscarpath, is_print = False):
        inf = {}
        
        infp = tkFile(poscarpath, 'r')
        if not infp or infp.fp is None:
            return None

        sample_name = infp.ReadLine().strip()
        if is_print:
            print("sample name: [{}]".format(sample_name))

        a0 = pfloat(infp.ReadLine())
        aij = []
        aij.append(infp.ReadLine().split())
        aij.append(infp.ReadLine().split())
        aij.append(infp.ReadLine().split())
        for i in range(3):
            for j in range(3):
                aij[i][j] = pfloat(aij[i][j])

        if is_print:
            print("")
            print("Base lattice parameter: {} A".format(a0))
            print("Lattice vectors:")
            for i in range(3):
                print("  {:16.12f}  {:16.12f}  {:16.12f}".format(*aij[i]))

        atomtypes = infp.ReadLine().split()
        natoms = infp.ReadLine().split()
        for i in range(len(natoms)):
            natoms[i] = pint(natoms[i])

        if is_print:
            print("")
            print("Atom types and numbers:")
            for i in range(len(atomtypes)):
                print("  {:>4}: {:4d}".format(atomtypes[i], natoms[i]))

        sitetypes = []
        for ia in range(len(atomtypes)):
            for n in range(natoms[ia]):
                sitetypes.append(atomtypes[ia])

#Selective dynamics
        line = infp.ReadLine().strip()
#Direct
        if line != 'Direct':
            line = infp.ReadLine()

        if is_print:
            print("")
            print("Atom positions:")

        pos = []    
        count = 0
        while 1:
            prevpos = infp.Tell()
            line = infp.ReadLine()
            if not line:
                break

            pos0 = line.split()
            if len(pos0) < 3:
                break

            try:
                pos0[0] = float(pos0[0])
            except:
                currpos = infp.Seek(prevpos)
                break

            pos.append([pfloat(pos0[0]), pfloat(pos0[1]), pfloat(pos0[2])])

            if is_print:
                print("  {:4}: ({:16.12f}, {:16.12f}, {:16.12f})"
                    .format(sitetypes[count], pos[count][0], pos[count][1], pos[count][2]))

            count += 1

        if is_print:
            print("")
            print("Build Crystal object:")

        cry = tkCrystal()
        cry.SetSampleName(sample_name)
        cry.SetCrystalName(sample_name)
        aij = a0 * np.array(aij)
#        print("aij_original=\n", aij)
        cry.SetLatticeVectors(aij)
#        print("aij_cry=\n", cry.LatticeVectors())
        latt = cry.LatticeParameters()
        for i in range(len(pos)):
            cry.AddAtomSite(name = sitetypes[i], pos = pos[i])

        cry.ExpandCoordinates()

#       cry.PrintInf()

        inf["SampleName"]  = sample_name
        inf["a0"]          = a0
        inf["aij"]         = np.array(a0) * aij
        inf["nAtomTypes"]  = len(atomtypes)
        inf["AtomTypes"]   = atomtypes
        inf["nAtoms"]      = natoms
        inf["crystal"]     = cry

        self.crystal = cry
        self.poscarinf = inf

        return cry

    def read_incar_inf(self, incarpath):
        fp = tkFile(incarpath, 'r')
        if not fp or fp.fp is None:
            return None

        inf = {}
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            
            line = tkre.Sub(r'[!#].*$', '', line)
#            print("line=", line)
            aa = line.split(';')
            for i in range(len(aa)):
                ret = fp.Search(r'(\S+)\s*[=:]\s*(\S.*)\s*', aa[i])
                if ret is None:
                    continue
#        print("key=", ret)
                inf[ret[1]] = pconv(ret[2])

        self.incarinf = inf
        
        return inf

    def read_oszicar_inf(self, oszicarpath):
        fp = tkFile(oszicarpath, 'r')
        if not fp or fp.fp is None:
            return None

        inf = {}
        ion_steps = []
        all_steps = []
        ion_step = 1
        while 1:
            line = fp.ReadLine()
            if not line or line == '':
                break
                
#            print("l: ", line, end = '')                    
            if 'DAV:' in line:
                aa = line.split()
                all_steps.append({ 'ion_step': ion_step, 'electron_step': pint(aa[1]), 'E': pfloat(aa[2]), 'dE': pfloat(aa[3])})
#                print("E=", aa[2])
            elif 'F=' in line:
                aa = line.split()
                ion_steps.append({ 'ion_step': aa[0], 'F': pfloat(aa[2]), 'E0': pfloat(aa[4]), 'dE': pfloat(aa[6])})
#                print("F=", aa[2], "  E=", aa[4])
                ion_step += 1
            elif 'ncg' in line:
                pass
            else:
                pass

        fp.Close()

        inf['ion_steps'] = ion_steps
        inf['all_steps'] = all_steps

        self.oszicarinf = inf

        return inf

    def read_outcar_inf(self, outcarpath):
        fp = tkFile(outcarpath, 'r')
        if not fp or fp.fp is None:
            return None

        inf = {}
        
# read input parameters
        line = fp.SkipTo(r'^\s*SYSTEM\s*=')
        if line:
            match = tkre.Search(r'SYSTEM\s*=\s*(\S.*?)\s*$', line)
            if match:
                inf['SYSTEM'] = match[1]

        line = fp.SkipTo("Startparameter for this run:")
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            if '----------' in line:
                break

            aa = line.split(';')
            for i in range(len(aa)):
                ret = tkre.Search(r'(\S+)\s*=\s*(\S+)', aa[i])
                if ret is None:
                    continue

                if len(ret) >= 3:
                    inf[ret[1]] = pconv(ret[2])
#                    print("ret:", ret[1], ret[2], pconv(ret[2]))

#    print("inf=", inf)
#    print("ISPIN=", inf["ISPIN"])

#        s = fp.SkipTo("E-fermi :")
#        s = tkre.Search(r"E-fermi :\s+([+\-\d\.eEdD]+)", line)
#        EF = pfloat(s[1])
        m = fp.SearchTo(r"E-fermi :\s*([+\-\d\.eDdD]+)", defval = "0.0")
        EF = pconv(m[1])
        inf["EF"]      = EF
        inf["E-fermi"] = EF

        fp.Rewind()
# Read average core potnetials
        Vcore = []
# average (electrostatic) potential at core
        line = fp.SkipTo("potential at core")
#  the test charge radii are     1.2793  0.7215
        fp.ReadLine()
#  (the norm of the test charge is              1.0000)
        fp.ReadLine()
#       1 -52.1673       2 -52.1683       3 -52.1600       4 -52.1667       5 -70.7270
#       6 -70.7333       7 -70.7334       8 -70.7355       9 -70.7342      10 -70.7302
        icount = 0
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            aa = line.split()
            if len(aa) <= 1:
                break

            for i in range(0, len(aa), 2):
                Vcore.append(pfloat(aa[i+1]))

        inf["Vcore"] = Vcore

        fp.Rewind()
# read ion step variation of converged total energy 
        count = 0
        energy = []
        while 1:
            line = fp.SkipTo("\\s*free  energy   TOTEN  =");
            if not line:
                break
        
            pos = fp.Tell()
            s = tkre.Search(r'free  energy   TOTEN  =\s*(\S+)', line)

            if pos > 0:
                fp.Seek(pos, 0)

            key = 'FreeEnergy{}'.format(count)
            energy.append(pfloat(s[1]))
#            print("E=", energy[count])
            count += 1
        inf['FreeEnergy'] = energy

# read final charges
        line = fp.SkipTo("\\s*total charge", 0)
        line = fp.SkipTo("# of ion     s       p       d       tot")
        line = fp.ReadLine()
        count = 0
        chargeinf = []
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            line = line.strip()
            if tkre.Match(r'^-----', line):
                break
            
            iIon, s, p, d, tot, = line.split()
#            print("i=", iIon, s, p, d, tot)
            chargeinf.append({
                     's'.format(iIon): pfloat(s)
                    ,'p'.format(iIon): pfloat(p)
                    ,'d'.format(iIon): pfloat(d)
                    ,'tot'.format(iIon): pfloat(tot)
                    })
            count += 1                

# Check EF after charge lines
# BZINTS: Fermi energy:  2.623512;********** electrons
        fp.Rewind()
        while 1:
            m = fp.SearchTo(r"BZINTS: Fermi energy:\s*([+\-\d\.eDdD]+)", defval = None)
            if m is None:
                break
            else:
                EF = pconv(m[1])
                inf["EF"]         = EF
                inf["EF(BZINTS)"] = EF

        inf['FinalCharges'] = chargeinf
        inf['nIon'] = count
#        if 'T' in sainf['LHFCALC']:
        if 'T' in inf.get('LHFCALC', 'F'):
            inf['IsHF'] = 1
        else:
            inf['IsHF'] = 0

        fp.Close()
        self.outcarinf = inf

        return inf

    def read_n_crystalstructures(self, outcarpath):
        fp = tkFile(outcarpath, 'r')
        n = 0
        if not fp:
            return 0

        while 1:
            line = fp.SkipTo('aborting loop because EDIFF is reached')
            if not line:
                break
        
            n += 1

        return n

    def read_next_crystalstructure(self, fp, template_cry, IsReduce01 = False):
        cry = copy.deepcopy(template_cry)
        inf = {}

        line = fp.SkipTo('aborting loop because EDIFF is reached')
        if not line:
            return None, None

        line = fp.SkipTo('VOLUME and BASIS-vectors')
        for i in range(4):
            fp.ReadLine()
        a1 = fp.ReadLine().split()
        a2 = fp.ReadLine().split()
        a3 = fp.ReadLine().split()
        a1[0], a1[1], a1[2] = pfloat(a1[0]), pfloat(a1[1]), pfloat(a1[2])
        a2[0], a2[1], a2[2] = pfloat(a2[0]), pfloat(a2[1]), pfloat(a2[2])
        a3[0], a3[1], a3[2] = pfloat(a3[0]), pfloat(a3[1]), pfloat(a3[2])
        cry.SetLatticeVectors([a1[:3], a2[:3], a3[:3]])
        latt = cry.LatticeParameters()

        line = fp.SkipTo('POSITION')
# skip ' ------...'
        line = fp.ReadLine()

        count = 0
        while 1:
            line = fp.ReadLine()
            if not line or '-----' in line:
                break
            x, y, z, fx, fy, fz, = line.split()
#           x = pfloat(x) / latt[0]
#            y = pfloat(y) / latt[1]
#            z = pfloat(z) / latt[2]

# Coordinate
            frx, fry, frz = cry.CartesianToFractional(pfloat(x), pfloat(y), pfloat(z))
            if IsReduce01:
                frx, fry, frz = Reduce01(frx), Reduce01(fry), Reduce01(frz)
#            print("")
#            print("***xyz=({},{},{})".format(x,y,z))
#            print("  frxyz=({},{},{})".format(frx,fry,frz))
#            x2,y2,z2 = cry.FractionalToCartesian(frx, fry, frz)
#            print("  2xyz=({},{},{})".format(x2,y2,z2))
#            exit()
# Force
            fx, fy, fz = cry.CartesianToFractional(pfloat(fx), pfloat(fy), pfloat(fz))
            cry.SetAtomSite(count, pos = [frx, fry, frz], 
                                   force = [pfloat(fx), pfloat(fy), pfloat(fz)])
            count += 1

        cry.ExpandCoordinates()

        line = fp.SkipTo("\\s*free  energy   TOTEN  =");
        ret  = fp.Search(r'free  energy   TOTEN  =\s*(\S+)', line)
        inf["TOTEN"] = pfloat(ret[1])
        
        pos = fp.Tell()

#       line = fp.SkipTo('MD-specific')
        line = fp.SkipTo('EKIN')
        if line:
            ret  = fp.Search(r'EKIN\s*=\s*([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["EKIN"] = pfloat(ret[1])
            line = fp.SkipTo('temperature')
            ret  = fp.Search(r'EKIN_LAT\s*=\s*([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["EKIN_LAT"] = pfloat(ret[1])
            ret  = fp.Search(r'temperature\s+([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["temperature"] = pfloat(ret[1])
            line = fp.SkipTo('ETOTAL')
            ret  = fp.Search(r'ETOTAL\s*=\s*([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["ETOTAL"] = pfloat(ret[1])
        else:
            fp.Seek(pos)

        return inf, cry

    def convert_kname(self, name):
        if len(name) <= 1:
            return name

        try:
            a = conv_dic[name.lower()]
            name = name.capitalize()
            name = '$\\{}$'.format(name)
        except:
            pass

        return name

    def read_kpoints_linemode(self, fp, inf, cry = None):
        IsEOF  = 0
        kp     = []
        kp_scf = []
        kp_all = []
        kpdic  = []
        ktot   = 0.0
        while 1:
            while 1:
                line1 = fp.ReadLine()
                if not line1:
                    IsEOF = 1
                    break

                line1 = line1.strip()
                if line1 == '':
                    next
                else:
                    break

            if IsEOF:
                break

            while 1:
                line2 = fp.ReadLine()
                if not line2:
                    IsEOF = 1
                    break

                line2 = line2.strip()
                if line2 == '':
                    next
                else:
                    break

            if IsEOF:
                break

#        print("line1: ", line1)
#        print("line2: ", line2)
            a1 = line1.split()
            n1 = len(a1)
            kx0, ky0,  kz0 = pfloat(a1[0]), pfloat(a1[1]), pfloat(a1[2])
            a2 = line2.split()
            n2 = len(a2)
            kx1, ky1,  kz1 = pfloat(a2[0]), pfloat(a2[1]), pfloat(a2[2])
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot += dk

            match = tkre.Search(r'[!#]\s*(\S+)', line1)
            if match:
                name0 = match[1]
            else:
                name0 = ''
            match = tkre.Search(r'[!#]\s*(\S+)', line2)
            if match:
                name1 = match[1]
            else:
                name1 = ''

            name0c = self.convert_kname(name0)
            name1c = self.convert_kname(name1)
            w = 0.0

            kinf = {"kx0": kx0, "ky0": ky0, "kz0": kz0, "kname0": name0, "kname0_conv": name0c,
                    "kx1": kx1, "ky1": ky1, "kz1": kz1, "kname1": name1, "kname1_conv": name1c,
                    "kw": w, "dk": dk, "ktot": ktot}
            kp.append(kinf)
            kp_all.append(kinf.copy())
            kpdic.append(kinf.copy())

        inf['nk']          = len(kp)
        inf['kpoints']     = kp
        inf['kpoints_scf'] = kp_scf
        inf['kpoints_all'] = kp_all
        inf['kpointsdic']  = kpdic

        return inf

    def read_kpoints_pointmode(self, fp, inf, cry = None):
        kp     = []
        kp_scf = []
        kp_all = []
        while 1:
            line1 = fp.ReadLine()
            if not line1:
                break

#            print("line1: ", line1)
            a1 = line1.split()
            n1 = len(a1)
            if n1 < 4:
                continue

            kx1, ky1,  kz1, w = pfloat(a1[0]), pfloat(a1[1]), pfloat(a1[2]), pfloat(a1[3])

            match = tkre.Search(r'[!#]\s*(\S+)', line1)
            if match:
                name1 = match[1]
            else:
                name1 = ''

            name1c = self.convert_kname(name1)

            dk   = None
            ktot = None
            kinf = {"kx0": kx1, "ky0": ky1, "kz0": kz1, "kname0": name1, "kname0_conv": name1c, 
                    "kx1": kx1, "ky1": ky1, "kz1": kz1, "kname1": name1, "kname1_conv": name1c, 
                    "kw": w, "dk": dk, "ktot": ktot}
            kp_all.append(kinf)
            if w == 0.0:
                kp.append(kinf.copy())
            else:
                kp_scf.append(kinf.copy())

            kx0, ky0, kz0 = kx1, ky1, kz1

        nk = len(kp)
        kpdic  = []
        ktot   = 0.0
        dkdic  = 0.0
        dkvec0 = None
        for ik in range(nk):
            ki = kp[ik]
            kx1, ky1,  kz1, w = ki["kx0"], ki["ky0"], ki["kz0"], ki["kw"]
            name1 = ki["kname0"]

            if ik == 0:
                kx0, ky0, kz0 = kx1, ky1, kz1

            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot  += dk
            dkdic += dk
            ki["dk"]   = dk
            ki["ktot"] = ktot

# add to kpdic
            if ik == 0:
                kpdic.append(ki.copy())
            elif ik == nk - 1:
                kpdic.append(ki.copy())
            else:
                kim = kp[ik-1]
                kip = kp[ik+1]
                kx0, ky0,  kz0, w = kim["kx0"], kim["ky0"], kim["kz0"], kim["kw"]
                kx2, ky2,  kz2, w = kip["kx0"], kip["ky0"], kip["kz0"], kip["kw"]
                r2   = (kx1 - kx0) * (kx1 - kx0) + (ky1 - ky0) * (ky1 - ky0) + (kz1 - kz0) * (kz1 - kz0)
                innp = (kx1 - kx0) * (kx2 - kx1) + (ky1 - ky0) * (ky2 - ky1) + (kz1 - kz0) * (kz2 - kz1)
                if r2 > 1.0e-6 and (name1 != '' or abs(innp) < 1.0e-6):
                    kidic = ki.copy()
                    kidic["dk"] = dkdic
                    kpdic.append(kidic)
                    dkdic = 0.0

            kx0, ky0, kz0 = kx1, ky1, kz1

        ktot   = 0.0
        for ik in range(len(kp_scf)):
            ki = kp_scf[ik]
            kx1, ky1,  kz1, w = ki["kx0"], ki["ky0"], ki["kz0"], ki["kw"]
            if ik == 0:
                kx0, ky0, kz0 = kx1, ky1, kz1
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot += dk
            ki["dk"]   = dk
            ki["ktot"] = ktot
            kx0, ky0, kz0 = kx1, ky1, kz1

        ktot   = 0.0
        for ik in range(len(kp_all)):
            ki = kp_all[ik]
            kx1, ky1,  kz1, w = ki["kx0"], ki["ky0"], ki["kz0"], ki["kw"]
            if ik == 0:
                kx0, ky0, kz0 = kx1, ky1, kz1
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot += dk
            ki["dk"]   = dk
            ki["ktot"] = ktot
            kx0, ky0, kz0 = kx1, ky1, kz1

        inf['nk']          = len(kp)
        inf['kpoints']     = kp
        inf['kpoints_scf'] = kp_scf
        inf['kpoints_all'] = kp_all
        inf['kpointsdic']  = kpdic

        return inf


    def read_kpoints(self, KPOINTS_path, cry = None):
        if cry is None:
            cry = self.crystal

        inf = {}

        fp = tkFile(KPOINTS_path, 'r')

        inf['title']      = fp.ReadLine().strip()
        inf['nk_target']  = pint(fp.ReadLine())
        mode              = fp.ReadLine().strip().upper()
        if mode[0] == 'L':
            inf['mode']       = mode
            inf['coord_mode'] = fp.ReadLine().strip()
            inf = self.read_kpoints_linemode(fp, inf, cry)
        else:
            inf['mode']       = mode
            inf['coord_mode'] = mode
            inf = self.read_kpoints_pointmode(fp, inf, cry)

        fp.Close()

        self.kpointsinf = inf

        return inf

    def read_eigenval(self, EIGENVAL_path, normalize_E = True, KPOINTS_inf = None, ISPIN = None, IsHF = None, EF = None, NELECT = None, cry = None):
        if cry is None:
            cry = self.crystal
        if NELECT is None:
            NELECT = self.outcarinf["NELECT"]
        if ISPIN is None:
            ISPIN = self.outcarinf["ISPIN"]
        if IsHF is None:
            IsHF = self.outcarinf["IsHF"]
        if EF is None:
            EF = self.outcarinf["EF"]

        try:
            kptinf = KPOINTS_inf['kpoints_all']
        except:
            kptinf = None
#        print("kpinf=", kpinf)

        inf = {}

        fp = tkFile(EIGENVAL_path, 'r')

#    4    4    1    1
        fp.ReadLine()
#  0.1236685E+02  0.3291500E-09  0.3291500E-09  0.5272308E-09  0.5000000E-15
        fp.ReadLine()
#  1.000000000000000E-004
        fp.ReadLine()
#  CAR 
        fp.ReadLine()
# ZnO [PAW_PBE54] (band)                  
        inf["title"] = fp.ReadLine().strip()
#     36    204     24
        aa = fp.ReadLine().split()
        inf["n"]       = pint(aa[0])
        inf["nk"]      = pint(aa[1])
        inf["nLevels"] = pint(aa[2])
        Elist     = []
        Elist_scf = []
        Elist_all = []
        kx0 = None
        ky0 = None
        kz0 = None
        for ik in range(inf["nk"]):
# next lines can be brank, to be skipped
            while 1:
                pos = fp.Tell()
                line = fp.ReadLine()
                if line == '':
                    break

#                print("line1=", line)
                if line.strip() != '':
                    fp.Seek(pos)
                    break

# 
#            fp.ReadLine()

#  0.0000000E+00  0.0000000E+00  0.0000000E+00  0.4901961E-02
            line = fp.ReadLine()
#            print("line2=", line)
            aa = line.split()
            kx = pfloat(aa[0])
            ky = pfloat(aa[1])
            kz = pfloat(aa[2])
            wk = pfloat(aa[3])

#    1      -16.924278   1.000000
            Eups   = []
            Edns   = []
            occups = []
            occdns = []
            for iL in range(inf["nLevels"]):
                line = fp.ReadLine()
#                print("line2=", line)
                aa = line.split()
                aa[1] = pfloat(aa[1])
                if len(aa) > 2:
                    aa[2] = pfloat(aa[2])
                else:
                    if ISPIN == 1:
                        aa.append(0.0)
                if ISPIN == 1:
                    if normalize_E:
                        Eups.append(aa[1] - EF)
                    else:
                        Eups.append(aa[1])
                    occups.append(aa[2])
                else:
                    if normalize_E:
                        Eups.append(aa[1] - EF)
                        Edns.append(aa[2] - EF)
                    else:
                        Eups.append(aa[1])
                        Edns.append(aa[2])
                    occups.append(pfloat(aa[3]))
                    occdns.append(pfloat(aa[4]))

            dk   = None
            ktot = None
            kinf = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]
            Elist_all.append(kinf)
            if IsHF:
                if not kptinf or kptinf[ik]['kw'] == 0.0:
                    Elist.append(kinf)
                else:
                    Elist_scf.append(kinf)
            else:
                Elist.append(kinf)

        fp.Close()

# Calculate k distances
        ktot = 0.0
        for ik in range(len(Elist)):
            kx, ky, kz, wk, _dk, _ktot, Eups, occups, Edns, occdns = Elist[ik]
            if kx0 is None:
                kx0, ky0, kz0 = kx, ky, kz
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx, ky, kz])
            ktot += dk
            kx0 = kx
            ky0 = ky
            kz0 = kz

            Elist[ik] = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]

        ktot = 0.0
        for ik in range(len(Elist_scf)):
            kx, ky, kz, wk, _dk, _ktot, Eups, occups, Edns, occdns = Elist_scf[ik]
            if kx0 is None:
                kx0, ky0, kz0 = kx, ky, kz
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx, ky, kz])
            ktot += dk
            kx0 = kx
            ky0 = ky
            kz0 = kz

            Elist_scf[ik] = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]

        ktot = 0.0
        for ik in range(len(Elist_all)):
            kx, ky, kz, wk, _dk, _ktot, Eups, occups, Edns, occdns = Elist_all[ik]
            if kx0 is None:
                kx0, ky0, kz0 = kx, ky, kz
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx, ky, kz])
            ktot += dk
            kx0 = kx
            ky0 = ky
            kz0 = kz

            Elist_all[ik] = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]


        inf['EList']     = Elist
        inf['EList_scf'] = Elist_scf
        inf['EList_all'] = Elist_all
        
        self.eigenvalinf = inf

        return inf

    def find_band_edges_from_dos(self, E, DOS, EF0 = 0.0, DOSth = 1.0e18):
        EV = 1.0e30
        EC = 1.0e30
        i0 = -1
        for i in range(len(E)):
#            print("i=", i, E[i], EF0)
            if E[i] >= EF0:
                i0 = i
                break

        for i in range(i0, 0, -1):
#            print("i2=", i, DOS[i], DOSth)
            if DOS[i] > DOSth:
                EV = E[i+1]
                break

        for i in range(i0, len(E), +1):
#            print("i3=", i, DOS[i], DOSth)
            if DOS[i] > DOSth:
                EC = E[i-1]
                break

        if EC < EV:
            EC = EF0
            EV = EF0

        return EV, EC

    def find_band_edges_from_eigenval(self, EF0 = 0.0, normalize_E = True, eigenvalinf = None, ISPIN = None):
        if ISPIN is None:
            if self.outcarinf:
                ISPIN = self.outcarinf["ISPIN"]
            elif self.incarinf:
                ISPIN = self.incarinf["ISPIN"]
            else:
                ISPIN = 1
        if EF0 is None:
            EF0 = 0.0
        if eigenvalinf is None:
            eigenvalinf = self.eigenvalinf
        EPS = 1.0e-3

        nk      = eigenvalinf["nk"]
        nLevels = eigenvalinf["nLevels"]
        EV = -1e100
        EC =  1e100
#        print("k points in EIGENVAL:")
#        print("nk=", nk)
#        print("nLevels=", nLevels)
        EHOMO = -1.0e10
        ELUMO =  1.0e10
        EList = eigenvalinf['EList']
        for i in range(len(EList)):
            el = EList[i]
            kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns = el
#            print("  ({:8.4f} {:8.4f} {:8.4f}) w={:8.4g}  {:8.4f} {:12.4f}"
#                    .format(kx, ky, kz, wk, dk, ktot))
            for il in range(nLevels):
#                print("  il={} E={} occ={}".format(il, Eups[il], occups[il]))
                if occups[il] > 0.0 and EHOMO < Eups[il]:
                    EHOMO = Eups[il]
                if occups[il] <= 0.0 and ELUMO > Eups[il]:
                    ELUMO = Eups[il]
                if Eups[il] < EF0 + EPS and EV < Eups[il]:
                    EV = Eups[il]
#                print("EV=>", i, EV+EF, kx, ky, kz)
                if Eups[il] > EF0 - EPS and EC > Eups[il]:
                    EC = Eups[il]

                if ISPIN == 2:
                    if occdns[il] > 0.0 and EHOMO < Edns[il]:
                        EHOMO = Edns[il]
                    if occdns[il] <= 0.0 and ELUMO > Edns[il]:
                        ELUMO = Edns[il]
                    if Edns[il] < EF0 + EPS and EV < Edns[il]:
                        EV = Edns[il]
#                        print("EV=>", i, EV+EF, kx, ky, kz)
                    if Edns[il] > EF0 - EPS and EC > Edns[il]:
                        EC = Edns[il]
#                print("EC=>", i, EC+EF, kx, ky, kz)
#            print("EV=", i, il, EV)

        if EC < EV:
            EC = EF0
            EV = EF0
            EHOMO = EF0
            ELUMO = EF0

        inf = { "EV": EV, "EC": EC, "Eg": EC - EV, "EF0": EF0, "EHOMO": EHOMO, "ELUMO": ELUMO }

        return inf

    def read_doscar(self, DOSCAR_path, cry = None, EF = None, normalize_E = True, unit = '', 
                IsSpinPolarized = None, IsNonCollinear = None, print_level = 1):
        if EF is None:
            EF = self.outcarinf["EF"]
        if cry is None:
            cry = self.crystal
        if IsSpinPolarized is None:
            IsSpinPolarized = self.outcarinf["ISPIN"]
        if IsNonCollinear is None:
            try:
                IsNonCollinear = cry.outcarinf["LCOLLINEAR"]
            except:
                IsNonCollinear = 0

        inf = {}

        fp = tkFile(DOSCAR_path, 'r')
        if not fp or fp.fp is None:
#            terminate("Error in read_average_core_potentials: Can not read [{}]".format(DOSCAR_path))
            return { "nE": 0, "E": [], "TotalDOSup": [], "TotalDOSdn": [], "Neup": [], "Nedn": [] }

#	if(!defined $IsNonCollinear) {
#		my $INCAR  = $this->GetINCARFileName($path);
#		my $pINCAR = $this->ReadINCARtoHash($INCAR);
#		$IsNonCollinear  = ($pINCAR->{LSORBIT} =~ /[1T]/i or $pINCAR->{LNONCOLLINEAR} =~ /[1T]/i)? 1 : 0;
#		$IsSpinPolarized = 0 if($IsNonCollinear);
#	}

        for i in range(5):
            line = fp.ReadLine()
        inf["SampleName"] = line.strip()
#    print("s=", inf["SampleName"])

        line = fp.ReadLine()
        aa = line.split()
        nData = pint(aa[2])
#    print("nData=", nData)

        Elist      = []
        tDOSuplist = []
        tDOSdnlist = []
        tDOSlist   = []
        Nelist     = []
        Neuplist   = []
        Nednlist   = []
        for i in range(nData):
            line = fp.ReadLine()
            if not line:
                break

            aa = line.split()
            if len(aa) == 0:
                break
            elif len(aa) <= 3:
                SpinPolarized = 0
                du = pfloat(aa[1])
                nu = pfloat(aa[2])
                dd = 0.0
                nd = 0.0
            else:
                SpinPolarized = 1
                du = pfloat(aa[1])
                dd = pfloat(aa[2])
                nu = 0.0
                nd = 0.0
            e = pfloat(aa[0])
            
            if normalize_E:
                Elist.append(e - EF)
            else:
                Elist.append(e)
            tDOSuplist.append(du)
            tDOSdnlist.append(dd)
            Neuplist.append(nu)
            Nednlist.append(du)
#        print("e=",e, du, nu)

        fp.Close()

        nE = len(Elist)
        
        if SpinPolarized == 1:
#            print("n=", len(tDOSuplist), len(tDOSdnlist))
            Neuplist = rieman_array_func(Elist, tDOSuplist)
            Nednlist = rieman_array_func(Elist, tDOSdnlist)

        if unit == '/cm3':
            Vcell = cry.Volume()
            for i in range(nE):
                k = 1.0 / (Vcell * 1.0e-24)
                tDOSuplist[i] *= k
                Neuplist[i]   *= k
                if SpinPolarized == 1:
                    tDOSdnlist[i] *= k
                    Neuplist[i]   *= k

        for i in range(nE):
            if SpinPolarized == 0:
                tDOSlist.append(tDOSuplist[i])
                Nelist.append(Neuplist[i])
            else:
                tDOSlist.append(tDOSuplist[i] + tDOSdnlist[i])
                Nelist.append(Neuplist[i] + Nednlist[i])

        inf["nE"]         = nE
        inf["E"]          = Elist
        inf["TotalDOS"]   = tDOSlist
        inf["TotalDOSup"] = tDOSuplist
        inf["TotalDOSdn"] = tDOSdnlist
        inf["Ne"]         = Nelist
        inf["Neup"]       = Neuplist
        inf["Nedn"]       = Nednlist

        return inf

    def read_DOS(file, cry = None, IsSpinPolarized = None, IsNonCollinear = None, EF = None, normalize_E = True, unit = '', print_level = 1):
        inf = {}
        if '.dat' in file:
            E_raw, dos_raw = np.genfromtxt(file, skip_header=1, usecols=(0,1), unpack=True)
            inf["E"]        = E_raw
            inf["nE"] = len(E_raw)
            inf["TotalDOS"] = dos_raw
        else:
            inf = self.read_doscar(file, cry = cry, EF = EF, normalize_E = normalize_E, unit = unit, 
                        IsSpinPolarized = IsSpinPolarized, IsNonCollinear = IsNonCollinear, print_level = print_level)

        return inf

    def read_files(self, file, file_types, EF = None, normalize_E = True, unit = '/cm3', data_for_bandedges = None, 
                    exit_by_error = False, print_level = 0, terminate = terminate): 
        base_path = self.getdir(file)

        INCAR_path    = self.get_INCAR(base_path)
        POSCAR_path   = self.get_POSCAR(base_path)
        POTCAR_path   = self.get_POTCAR(base_path)
        CONTCAR_path  = self.get_CONTCAR(base_path)
        OUTCAR_path   = self.get_OUTCAR(base_path)
        EIGENVAL_path = self.get_VASPPath(base_path, 'EIGENVAL')
        DOSCAR_path   = self.get_VASPPath(base_path, 'DOSCAR')
        if print_level >= 1:
            print("")
            print("CAR dir   : ", base_path)
            print("  INCAR   : ", INCAR_path)
            print("  POSCAR  : ", POSCAR_path)
            print("  POTCAR  : ", POTCAR_path)
            print("  CONTCAR : ", CONTCAR_path)
            print("  OUTCAR  : ", OUTCAR_path)
            print("  EIGENVAL: ", EIGENVAL_path)
            print("  DOSCAR  : ", DOSCAR_path)

        inf = {}
        if 'INCAR' in file_types:
            if print_level >= 1:
                print("*** Read INCAR from [{}]".format(INCAR_path))
            inf["INCAR"] = self.read_incar_inf(INCAR_path)
            if exit_by_error and inf["INCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(INCAR_path))

        if 'POTCAR' in file_types:
            if print_level >= 1:
                print("*** Read POTCAR from [{}]".format(POTCAR_path))
            inf["POTCAR"] = self.read_potcar_inf(POTCAR_path)
            if exit_by_error and inf["POTCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(POTCAR_path))

        if 'POSCAR' in file_types:
            if print_level >= 1:
                print("*** Read crystal structure from [{}]".format(POSCAR_path))
            inf["POSCAR"] = self.read_poscar(POSCAR_path)
            if exit_by_error and inf["POSCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(POSCAR_path))

        if 'CONTCAR' in file_types:
            if print_level >= 1:
                print("*** Read crystal structure from [{}]".format(CONTCAR_path))
            inf["CONTCAR"] = self.read_poscar(CONTCAR_path)
            if exit_by_error and inf["POSCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(CONTCAR_path))

        if 'OUTCAR' in file_types:
            if print_level >= 1:
                print("*** Read OUTCAR from [{}]".format(OUTCAR_path))
            inf["OUTCAR"] = self.read_outcar_inf(OUTCAR_path)
            if exit_by_error and inf["OUTCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(OUTCAR_path))

        if 'DOSCAR' in file_types:
            if print_level >= 1:
                print("*** Read DOSCAR from [{}]".format(DOSCAR_path))
            inf["DOSCAR"] = self.read_doscar(DOSCAR_path, EF = EF, normalize_E = normalize_E, unit = unit, print_level = print_level)
            if exit_by_error and inf["DOSCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(DOSCAR_path))

        if 'EIGENVAL' in file_types:
            if print_level >= 1:
                print("*** Read EIGENVAL from [{}]".format(EIGENVAL_path))
            inf["EIGENVAL"] = self.read_eigenval(EIGENVAL_path, normalize_E = normalize_E)
            if exit_by_error and inf["EIGENVAL"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(EIGENVAL_path))

        return inf

