import os
import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt
from scipy.optimize import minimize
import pandas as pd
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score
from matplotlib import pyplot as plt
import matplotlib.widgets as wg


from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg, pconv_by_type
#analyze_varstr, save_csv
from tklib.tkutils import is_exist, is_file, is_dir
from tklib.tkinifile import tkIniFile
from tklib.tkparams import tkParams
from tklib.tkgraphic.tkplotevent import tkPlotEvent

from tklib.tkvariousdata import tkVariousData
#from tklib.tksci.tkmatrix import make_matrix1
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius, log10
from tklib.tksci.tkoptimize import mlsq_general


'''
Fitting and interactive matplotlib helper object library
'''


class tkFit_object(tkParams):
    def __init__(self, parameter_file = None, app = None, 
                    tol = 1.0e-5, nmaxiter = 100,
                    print_interval = 1, plot_interval = 1, **args):
        super(tkParams, self).__init__(**args)
#        super(tkParams, self).__init__(parameter_file, app, args)
#        self.update(**args)

        self.plot_event = tkPlotEvent(plt)
        self.varname  = []
        self.unit     = []
        self.pk       = []
        self.optid    = []
        self.kpenalty = []
        self.kmin     = []
        self.kmax     = []
        
        self.tol = tol
        self.print_interval = print_interval
        self.nmaxiter       = nmaxiter
        self.plot_interval  = plot_interval
        self.stop_flag = False

        self.x    = None
        self.y    = None
        self.func = None
        self.iter = 0
        self.f    = None

    def get_value(self, target_key, keys, values, def_val = None):
        for i in range(len(keys)):
            key = keys[i]
            if key == target_key:
                return values[i]
        return def_val

    def read_parameters(self, prmfile, heading = None, section = None, 
                        keys = None, values = None, optid = None, kmin = None, kmax = None, kpenalty = None,
                        ignore_keys = []):
        if heading is not None:
            print(heading)
        if keys is None:
            keys = self.varname
        if values is None:
            values = self.pk
        if optid is None:
            optid = self.optid
        if kmin is None:
            kmin = self.kmin
        if kmax is None:
            kmax = self.kmax
        if kpenalty is None:
            kpenalty = self.kpenalty

        p = tkParams()
        p.read_parameters(prmfile, section = section)
#        p.print_parameters(heading = "p")
        for i in range(len(keys)):
            key = keys[i]
            if key in ignore_keys:
                continue

            v = p.__dict__.get(key, None)
            if v is None:
                continue

#            print("type=", i, keys[i], type(values[i]))
            t = type(values[i])
            if ':' in v:
                aa = v.split(':')
                n = len(aa)
#                print("aa=", key, v, aa)
#                values[i] = pfloat(aa[0], defval = values[i])
                values[i] = pconv_by_type(aa[0], type = type(values[i]), defval = values[i], strict = True)
                if n >= 2:
                    optid[i] = pint(aa[1])
                if n >= 3 and len(kmin) >= i+1:
                    kmin[i] = pfloat(aa[2])
                if n >= 4 and len(kmax) >= i+1:
                    kmax[i] = pfloat(aa[3])
                if n >= 5 and len(kpenalty) >= i+1:
                    kpenalty[i] = pfloat(aa[4])
#                print("v=", i, values[i], optid[i])
            else:
#                values[i] = pfloat(v, defval = values[i])
                values[i] = pconv_by_type(v, type = type(values[i]), defval = values[i], strict = True)
#                print("v=", i, values[i])

    def save_parameters(self, prmfile, heading = None, section = None, 
                            keys = None, values = None, optid = None, kmin = None, kmax = None, kpenalty = None):
        if heading is not None:
            print(heading)
        if keys is None:
            keys = self.varname
        if values is None:
            values = self.pk
        if optid is None:
            optid = self.optid
        if kmin is None:
            kmin = self.kmin
        if kmax is None:
            kmax = self.kmax
        if kpenalty is None:
            kpenalty = self.kpenalty

        ini = tkIniFile(path = prmfile)

        for i in range(len(keys)):
            key = keys[i]
            if len(kpenalty) > 0:
                value = f"{values[i]}:{optid[i]}:{kmin[i]}:{kmax[i]}:{kpenalty[i]}"
            else:
                value = f"{values[i]}:{optid[i]}"
            ini.write_string(section = section, key = key, value = value, is_print = False)

    def extract_parameters(self, pk = None, optid = None):
        if pk is None:
            pk = self.pk
        if optid is None:
            optid = self.optid
        
        optpk = []
        for i in range(len(pk)):
            if optid[i]:
                optpk.append(pk[i])
        return optpk

    def recover_parameters(self, optpk, allpk = None, optid = None, set_member = True):
        if allpk is None:
            allpk = self.pk
        if optid is None:
            optid = self.optid

        pk = []
        c = 0
        for i in range(len(allpk)):
            if optid[i]:
                pk.append(optpk[c])
                c += 1
            else:
                pk.append(allpk[i])

        if set_member:
            self.pk = pk

        return pk

    def print_variables(self, heading = None, varname = None, unit = None, pk = None, dx = None, optid = None, 
                        kmin = None, kmax = None, kpenalty = None, f = None):
        if heading is not None:
            print(heading)
        if varname is None:
            varname = self.varname
        if unit is None:
            unit = self.unit
        if pk is None:
            pk = self.pk
        if optid is None:
            optid = self.optid
        if kmin is None:
            kmin = self.kmin
        if kmax is None:
            kmax = self.kmax
        if kpenalty is None:
            kpenalty = self.kpenalty

        np = len(kpenalty)
        for i in range(len(pk)):
            if np > 0:
                print(f"   {varname[i]:>10}={pk[i]:12.6g} {unit[i]:10} (id={optid[i]}) "
                    + f"penality: {kpenalty[i]:8.3g} * ({kmin[i]:g} - {kmax[i]:g})")
            else:
                print(f"   {varname[i]:>10}={pk[i]:12.6g} {unit[i]:10} (id={optid[i]})")
        if f is not None:
            print("  f=", f)

    def read_data(self, infile, xlabel = 0, ylabel = 1, xmin = None, xmax = None, usage = None):
        print("")
        print(f"Read [{infile}]")
        datafile = tkVariousData(infile)
        labels, datalist = datafile.Read_minimum_matrix(close_fp = True, force_numeric = False, usage = usage)
        xlabel, xin = datafile.FindDataArray(xlabel, flag = 'i')
        ylabel, yin = datafile.FindDataArray(ylabel, flag = 'i')

        self.datafile = datafile
        self.labels   = labels
        self.datalist = datalist
        self.xlabel   = xlabel
        self.ylabel   = ylabel
        
        self.ndata_all = len(xin)
        print("ndata_all=", self.ndata_all)
        self.x = []
        self.y = []
        self.included_index = []
        for i in range(self.ndata_all):
            if xmin is not None and xmin > xin[i]:
                continue
            if xmax is not None and xmax < xin[i]:
                continue
            
            self.x.append(xin[i])
            self.y.append(yin[i])
            self.included_index.append(i)

        self.ndata = len(self.x)
        self.index = range(self.ndata)

    def print_data(self, heading = '', yini = None, yfin = None):
        nx    = len(self.x_list)
        ndata = len(self.x_list[0])

        print(f"{'i':>4}:{'sn':<4} ", end = '')
        for j in range(nx):
            print(f"{self.xlabels[j]:>10} ", end = '')
        if yfin is None:
            print(f"{'y(input)':>10} {'y(ini)':>10}")
        else:
            print(f"{'y(input)':>10} {'y(ini)':>10} {'y(fin)':>10}")
        for i in range(ndata):
            print(f"{self.index[i]:>4}:{self.included_index[i]:<4}: ", end = '')
            for j in range(nx):
                print(f"{self.x_list[j][i]:10.4g} ", end = '')
            if yfin is None:
                print(f"{self.y[i]:10.4g} {yini[i]:10.4g}")
            else:
                print(f"{self.y[i]:10.4g} {yini[i]:10.4g} {yfin[i]:10.4g}")

    def to_excel(self, outfile, labels, data_list):
        nx = len(data_list)
        max_ndata = 0
        for d in data_list:
            ndata = len(d)
            if max_ndata < ndata:
                max_ndata = ndata
        for i in range(nx):
            ndata = len(data_list[i])
            if ndata < max_ndata:
                if type(data_list[i]) is not list:
                    data_list[i] = data_list[i].tolist()
                for j in range(max_ndata - ndata):
                    data_list[i].append(None)

        df = pd.DataFrame(np.array(data_list).T, columns = labels)
        df.to_excel(outfile, index = False)

        if not os.path.exists(outfile):
            print("")
            print(f"Error in tkFit.to_excel(): Could not write to [{outfile}].")
            return False

        return True

    def print_scores(self, y1, y2, heading = None):
        if heading is not None:
            print(heading)

        y1_mean = np.mean(y1)
        y2_mean = np.mean(y2)
        y1_var  = np.var(y1)
        y2_var  = np.var(y2)
        y1_std0 = np.std(y1)
        y2_std0 = np.std(y2)
        y1_std1 = np.std(y1, ddof = 1)
        y2_std1 = np.std(y2, ddof = 1)
        MAE  = mean_absolute_error(y1, y2)
        MSE  = mean_squared_error(y1, y2)
        RMSE = sqrt(MSE)
        R2 = r2_score(y1, y2)
        sxx = np.dot(y1, y1)
        syy = np.dot(y2, y2)
        sxy = np.dot(y1, y2)
        rcorr = sxy / sqrt(sxx * syy)

        print(f"  Mean values = <y1> = sum(y1) / n                 : {y1_mean:12.4g}")
        print(f"                <y2> = sum(y2) / n                 : {y2_mean:12.4g}")
        print(f"  Variance    = sum((y1 - <y1>)^2) / n             : {y1_var:12.4g}")
        print(f"              = sum((y2 - <y2>)^2) / n             : {y2_var:12.4g}")
        print(f"  Standard deviation = sqrt(Variance(y1))          : {y1_std0:12.4g}")
        print(f"                       sqrt(Variance(y2))          : {y2_std0:12.4g}")
        print(f"  Sample std  = sqrt(sum(y1 - <y1>)^2 / (n-1)      : {y1_std1:12.4g}")
        print(f"                sqrt(sum(y1 - <y1>)^2 / (n-1)      : {y2_std1:12.4g}")
        print(f"  MAE  (mean absolute error) = sum(|y1 - y2|) / n  : {MAE:12.4g}")
        print(f"  MSE  (mean squared error)  = sum((y1 - y2)^2) / n: {MSE:12.4g}")
        print(f"  RMSE (root MSE)            = sqrt(MSE)           : {RMSE:12.4g}")
        print(f"  R^2  (coefficient of determnation)")
        print(f"           = 1 - sum((y1 - y2)^2 / sum(y1 - <y2>)^2: {R2:12.4g}")
        print(f"  r    (correlation coefficient)                   : {rcorr:12.4g}")

    def initialize_plot(self, fig = None, use_pause = 0.00001, 
                        button_region = [0.15, 0.95, 0.10, 0.03], plot_region = [0.92, 0.15],
                        on_clicked = None):
        if on_clicked is None:
            on_clicked = self.button_click

        self.plt = plt
        self.fig = fig
        self.use_pause = use_pause

        if self.fig is None:
            print("")
            print("Warning: figure object is not given.")
            print("    Mouse click event is not bound")
            print("")
        else:
            self.plot_event.register_event(self.fig, event = "button_press_event", 
                        callback = lambda event: self.plot_event.onclick(event))

        self.button_region = button_region
        self.plot_region   = plot_region
        self.button_region = button_region
        self.ax_button     = plt.axes(button_region)
        self.stop_button   = wg.Button(self.ax_button, 'stop', color = '#f8e58c', hovercolor = '#38b48b')
        self.stop_button.on_clicked(self.button_click)

    def layout(self, show = False):
        self.plt.tight_layout()
        self.plt.subplots_adjust(top = self.plot_region[0], bottom = self.plot_region[1])
        
        if show:
            if use_pause is not None:
                self.plt.pause(use_pause)
            else:
                self.plt.show()

