import openpyxl
from math import exp, sqrt, log
import numpy as np
from scipy import optimize          # newton関数はscipy.optimizeモジュールに入っている
from scipy.interpolate import interp1d


from tklib.tkobject import tkObject
from tklib.tkutils import terminate, pfloat, pint, safe_getelement, validate_error, split_optstr
from tklib.tkinifile import tkIniFile
from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me
from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tktransport.tkTransport import read_datafile
from tklib.tktransport.tkTransport import LorentzNumber_FEA, FermiIntegral_x2
from tklib.tktransport.tkTransport import FermiIntegral_fast as FermiIntegral
from tklib.tktransport.tkTransport import rieman, integrate_Simpson, integrate_Simpson_list, fe, fh
from tklib.tktransport.tkTransport import meff2NC_FEA, NC2meff_FEA, meff2DC0_FEA


kpi05   = 2.0 / sqrt(pi)
kBe     = kB / e
kBe2    = kB * kB / e / e
ksigma0 = 16.0 * pi * e * e / 3.0 / h**3 * me
kmu0    = 2.0 * e / 3.0 / sqrt(2.0 * me)
ktau0   = 2.0 / 3.0 * sqrt(me * 0.5)
kS2     = kB / e * pi * pi / 3.0 * 1.5
kS3     = 8.0 / 3.0 * pi * pi * kB * kB / e / h / h * pow(pi / 3.0, 2.0/3.0) 
kkappa0 = 16.0 * pi * kB / 3.0 / h**3 * me


class tkDOS_FEA_single_carrier(tkObject):
    def __init__(self):
        super(tkObject, self).__init__()

    def __del__(self):
#        super(tkObject, self).__del__()
        pass

    def __str__(self):
        return self.ClassPath()


    def set(self, **args):
        self.update(args)

    def read_excel_Hall2(self, infile, usage = None):
        datafile = tkVariousData(infile)
        labels, datalist = datafile.Read_minimum_matrix(close_fp = True, usage = usage)
        label_T, xT           = datafile.FindDataArray(r'^T[\s|\(|\[|$]', flag = 'i')
        label_sigma, ysigma   = datafile.FindDataArray(r'^sigma[$|\S]?.*$', flag = 'i')
        label_N, yN           = datafile.FindDataArray(r'^N[$|\S]?.*$', flag = 'i')
        label_mu, ymu         = datafile.FindDataArray(r'^mu[$|\s|\(\[]?.*$', flag = 'i')
        label_RH, yRH         = datafile.FindDataArray(r'^RH[$|\s|\(\[]?.*$', flag = 'i')
        if yRH is None:
            label_RH = 'RH(cm3/C)'
            yRH = [1.0 / e / yN[i] for i in range(len(yN))]

        self.header = [label_T, label_N, label_mu, label_sigma, label_RH]
        self.xT  = xT
        self.yN  = yN
        self.ymu = ymu
        self.ys  = ysigma
        self.yRH = yRH
        self.label_T     = label_T
        self.label_N     = label_N
        self.label_mu    = label_mu
        self.label_sigma = label_sigma
        self.label_RH    = label_RH

        return self.header, xT, yN, ymu, ysigma, yRH

    def read_excel_Hall(self, infile, usage = None):
        self.read_excel_Hall2(infile, usage)
        
        return self.header, self.xT, self.yN, self.ymu, self.ys

    def set_misc_parameters(self, params):
        self.Einteg0             = params.get("Einteg0",             self.Einteg0)
        self.Einteg1             = params.get("Einteg1",             self.Einteg1)
        self.nrange              = params.get("nrange",              self.nrange)
        self.print_level         = params.get("print_level",         self.print_level)
        self.EF0                 = params.get("EF0",                 self.EF0)
        self.nEF                 = params.get("nEF",                 self.nEF)
        self.Estep               = params.get("Estep",               self.Estep)
        self.dE_bisec            = params.get("dE_bisec",            self.dE_bisec)
        self.eps_bisec           = params.get("eps_bisec",           self.eps_bisec)
        self.nmaxiter_bisec      = params.get("nmaxiter_bisec",      self.nmaxiter_bisec)
        self.iprintiterval_bisec = params.get("iprintiterval_bisec", self.iprintiterval_bisec)

    def parameter_list(self):
        list = []
        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            list.append(self.get(key))
        return list

    def set_parameters(self, ai):
        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            self.set_attribute(key, ai[i])

    def read_parameters(self, path, optid, mobility):
        ini = tkIniFile()
        inf = ini.ReadAll(path, AddSection = 0)
        if inf is None:
            return

        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            str = inf.get(key, None)
            if str is None:
                continue
            val, id = split_optstr(str)
            self.set_attribute(key, val)
            if id is not None:
                optid[i] = id

        mobility.charge  = pfloat(safe_getelement(inf, "charge", mobility.charge))
        mobility.rfac    = pfloat(safe_getelement(inf, "rfac", mobility.rfac))
        self.print_level = pfloat(safe_getelement(inf, "print_level", self.print_level))

    def save_parameters(self, path, optid, otherparams = None, parameters_section = 'Parameters', other_section = 'OtherParameters'):
        ini = tkIniFile(path)
        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            val = self.get(key, None)
            if val is not None:
                ini.WriteString(parameters_section, key, "{}:{}".format(val, optid[i]))

        if otherparams is not None:
            for key in otherparams.keys():
                val = otherparams[key]
                if val is not None:
                    ini.WriteString(other_section, key, val)

    def print_parameters(self, varname, varunit, ai, optid):
        for i in range(len(self.varkeys)):
            key = self.varkeys[i]
            val = self.get(key, None)
            print("  {:10s}: {:14.8g}".format(key, val))

    def meff2NC_FEA(self, meff, T):
        val = meff2NC_FEA(meff, T)
        return val

    def NC2meff_FEA(self, NC, T):
        return NC2meff_FEA(NC, T)

    def meff2DC0_FEA(self, meff, T):
        return meff2DC0_FEA(meff, T)

# define the DOS function
    def Dv(self, E):
        if self.EV < E:
            return 0.0
        else:
            return self.DV0 * sqrt(self.EV - E)

    def Dc(self, E):
        if E < self.EC:
            return 0.0
        else:
            return self.DC0 * sqrt(E - self.EC)

    def DOS(self, E):
        if E < self.EV:
            return self.DV0 * sqrt(self.EV - E)
        elif self.EC < E:
            return self.DC0 * sqrt(E - self.EC)
        return 0.0

# Define the function to be integrated
    def DOSfe(self, E, T, EF):
        return self.DOS(E) * fe(E, T, EF)

    def DOSfh(self, E, T, EF):
        return self.DOS(E) * fh(E, T, EF)

# Calculate the number of electrons from E0 to E1 with the Fermi level EF at T
    def cal_electrondensity(self, EF, T):
        F05 = FermiIntegral((EF - self.EC) * e / kB / T, 0.5)
        return self.NC * kpi05 * F05                                                # cm^-3

    def Ne(self, T, EF, E0, E1, Estep):
        x  = (EF - self.EC) * e / kB / T
        F05 = FermiIntegral(x, 0.5)
        N = self.NC * kpi05 * F05                                                # cm^-3
        return N

        """
        ET = kB * T / e
        Estep = min([ET * 0.01, Estep])
        N = integrate_Simpson(lambda E: self.DOSfe(E, T, EF), E0, E1, Estep)
        return N
        """

    def Nh(self, T, EF, E0, E1, Estep):
        x  = (self.EV - EF) * e / kB / T
        F05 = FermiIntegral(x, 0.5)
        N = self.NV * kpi05 * F05                                                # cm^-3
        return N

        """
        ET = kB * T / e
        Estep = min([ET * 0.01, Estep])
        N = integrate_Simpson(lambda E: self.DOSfh(E, T, EF), E0, E1, Estep)
        return N
        """

# ionized donor density
    def NDp(self, EF, T):
        return self.ND * fh(self.ED, T, EF)
#        return self.ND * (1.0 - fe(self.ED, T, EF))

# ionized acceptor density
    def NAm(self, EF, T):
        return self.NA * fe(self.EA, T, EF)

# Calculate EF base on 
# the charge neutrality (electron number) condition, Ne(T, EF, dE) - N = 0,
# using the Newton method with the initial value EF0
    def FindEF_metal(self, T, EF0, E0, E1, Estep, totNe):
        EF = optimize.newton(self.Ne(T, EF, E0, E1, Estep) - totNe, self.EF0)
        return EF

    def FindEF_semi(self, T, EFmin, EFmax, E0h, E1h, E0e, E1e, Estep, print_level = 1):
# まず、EFmin,EFmaxにおけるΔQを計算し、それぞれが正・負あるいは負・生となっていることを確認する    
        if print_level >= 2:
            print("E range:", E0h, E1e)

        while 1:
            Nhmin, Nemin, NAmmin, NDpmin, RHmin, Nsmin = self.cal_carrier_densities(T, EFmin, E0h, E1h, E0e, E1e, Estep)
            dQmin  = Nemin + NAmmin - Nhmin - NDpmin
            Nhmax, Nemax, NAmmax, NDpmax, RHmax, Nsmax = self.cal_carrier_densities(T, EFmax, E0h, E1h, E0e, E1e, Estep)
            dQmax = Nemax + NAmmax - Nhmax - NDpmax
            iT = 0
            if print_level >= 2:
                print("        min (EF,Ne,Nh,NDp,NAm,dQ)=({:6.4f},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                            .format(iT, EFmin, Nemin, Nhmin, NDpmin, NAmmin, dQmin))
                print("        max (EF,Ne,Nh,NDp,NAm,dQ)=({:6.4f},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                            .format(EFmax, Nemax, Nhmax, NDpmax, NAmmax, dQmax))
#        dQmin = Ne(EFmin, T) + NAm(EFmin, T) - Nh(EFmin, T) - NDp(EFmin, T)
#        dQmax = Ne(EFmax, T) + NAm(EFmax, T) - Nh(EFmax, T) - NDp(EFmax, T)
#        print("  EFmin = {:12.8f}  dQmin = {:12.4g}".format(EFmin, dQmin))
#        print("  EFmax = {:12.8f}  dQmax = {:12.4g}".format(EFmax, dQmax))
            if print_level >= 2:
                print("    dQ=", dQmin, dQmax)
                print("    dQmin*dQmax=", dQmin * dQmax)
            if dQmin * dQmax > 0.0:
                if dQmin > 0.0:
                    EFmin -= 0.1
                elif dQmax < 0.0:
                    EFmax += 0.1
                if print_level >= 3:
                    print("Warning in tkDOS_FEA.FindEF_semi(): Initial Emin {:8.3g} and Emax {:8.3g} should be chosen as dQmin * dQmax < 0"
                                .format(EFmin, EFmax))
                    print(f"  EF range is expanded to [{EFmin}] - [{EFmax}] eV")
#                return None, None
            else:
                break

# 2分法開始
        for i in range(self.nmaxiter_bisec):
            EFhalf = (EFmin + EFmax) / 2.0
            Nhh, Neh, NAmh, NDph, RHh, Nsh = self.cal_carrier_densities(T, EFhalf, E0h, E1h, E0e, E1e, Estep)
            dQhalf = Neh + NAmh - Nhh - NDph
            if print_level >= 2:
                print("  iter {:03d}: half (EF,Ne,Nh,NDp,NAm,dQ)=({:6.4f},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                        .format(i, EFhalf, Neh, Nhh, NDph, NAmh, dQhalf))
# EFの精度がepsより小さくなったら計算終了
            if abs(EFmin - EFhalf) < self.eps_bisec and abs(EFmax - EFhalf) < self.eps_bisec:
#                    print("  Success: Convergence reached at EF = {}".format(EFhalf))
                    break
            if dQmin * dQhalf < 0.0:
                EFmax = EFhalf
                dQmax = dQhalf
            else:
                EFmin = EFhalf
                dQmin = dQhalf
        else:
            print("  Failed: Convergence did not reach")
            return EFhalf, dQhalf

        return EFhalf, dQhalf

    def EF_from_electrondensity(self, Ne, T, EF0 = 0.0, dEF = 0.01, dump = 0.0, epsEF = 1.0e-5, maxiter = 100):
        lnNe      = log(Ne)
        EFprev    = EF0
        dlnNeprev = log(self.cal_electrondensity(EF0, T)) - lnNe
        EF        = EFprev + dEF
#        print("Ne={:g}".format(Ne))
#        print("EF0=", EF0)
        for i in range(maxiter):
            n = self.cal_electrondensity(EF, T)
            dlnNe  = log(n) - lnNe
#            print("EF={:8.4g}, n={:12.6g} d_ln(N)={:12.6g}".format(EF, n, dlnNe))
            diffNe = (dlnNe - dlnNeprev) / (EF - EFprev)
#            print("d ln(N) / dEF=", diffNe)
            if diffNe < 0.0:
                k = -dump + diffNe
            else:
                k = dump + diffNe
            diffEF = -dlnNe / k
            if(abs(diffEF) < epsEF):
                break

            EFprev = EF
            EF += diffEF
            dlnNeprev = dlnNe
#            print("EF1=", EF, diffEF)
        else:
            print("")
            print("Warning in tkDOS_FEA.EF_from_electrondensity(): Not converged diff={} eps={}".format(diffEF, epsEF))
            return EF, diffEF, False

        return EF, diffEF, True

    def EF_from_electronSeebeck(self, Starget, T, mobility, EF0 = 0.0, dEF = 0.01, dump = 0.0, epsEF = 1.0e-5, maxiter = 100):
        EFprev= EF0
        dSprev = self.cal_electronSeebeck(EF0, T, mobility) - Starget
        EF     = EFprev + dEF
        for i in range(maxiter):
            S = self.cal_electronSeebeck(EF, T, mobility)
            dS  = S - Starget
            diffS = (dS - dSprev) / (EF - EFprev)
            if diffS < 0.0:
                k = -dump + diffS
            else:
                k = dump + diffS
            diffEF = -dS / k
            if(abs(diffEF) < epsEF):
                break

            EFprev = EF
            EF += diffEF
            dSprev = dS
        else:
            print("")
            print("Warning in tkDOS_FEA.EF_from_electronSeebeck(): Not converged diff={} eps={}".format(diffEF, epsEF))
            return EF, diffEF, False

        return EF, diffEF, True

    def cal_electronSeebeck(self, EF, T, mobility):
        x  = (EF - self.EC) * e / kB / T

        Fr  = FermiIntegral(x, mobility.rfac)
        Fr1 = FermiIntegral(x, mobility.rfac + 1.0)
        S = kB / e * ( (mobility.rfac + 2.0) / (mobility.rfac + 1.0) * Fr1 / Fr - x) * 1.0e6      # uV/K

        return S

    def cal_transport_properteis(self, T, EF, mobility, klatt = 0.0, validate_error_str = None, charge = None):
        if charge is None:
            charge = mobility.charge

        x  = EF * e / kB / T

        mobility.set_scattering_parameters()
        self.NC  = meff2NC_FEA(self.meeff, T)
        self.DC0 = meff2DC0_FEA(self.meeff, T)

        ksigma = ksigma0 * mobility.l0norm * self.meeff
        kmu    = kmu0 * mobility.l0norm / sqrt(self.meeff) 
        kkappa = kkappa0 * mobility.l0norm * self.meeff
        ktau   = ktau0 * mobility.l0norm * (mobility.rfac + 1.0) * sqrt(self.meeff) 

        kBT = kB * T

        F05 = FermiIntegral(x, 0.5)
        Fr  = FermiIntegral(x, mobility.rfac)
        Fr1 = FermiIntegral(x, mobility.rfac + 1.0)
        Fr2 = FermiIntegral(x, mobility.rfac + 2.0)
# Carrier density, conductivity, mobility
        n     = self.NC * kpi05 * F05                                                # cm^-3
        sigma = ksigma * pow(kBT, mobility.rfac + 1.0) * (mobility.rfac + 1.0) * Fr / 100.0   # S/cm
        mu    = sigma / e / n                                                   # cm2/Vs
#        print("mu={}/{} cm2/Vs".format(mu, mu2))
        if validate_error_str is not None and validate_error_str != '':
            mu2 = kmu * pow(kBT, mobility.rfac - 0.5) * (mobility.rfac + 1.0) * Fr / F05 *1.0e4 # cm2/Vs
            validate_error(mu, mu2, 1.0e-10, "{} for mu: ".format(validate_error_str))

# Relaxation time average
        tau_avg = mu * 1.0e-4 * self.meeff * me / e * 1.0e15                          # fs
#        print("tau=", tau, tau_avg, tau_avg2)
#        print("<tau>={}/{} fs".format(tau_avg, tau_avg2))
        if validate_error_str is not None and validate_error_str != '':
            tau_avg2 = ktau * pow(kBT, mobility.rfac - 0.5) * Fr / F05 * 1.0e15
            validate_error(tau_avg, tau_avg2, 1.0e-10, "{} for <tau>: ".format(validate_error_str))

        S = kBe * ( (mobility.rfac + 2.0) / (mobility.rfac + 1.0) * Fr1 / Fr - x) * 1.0e6      # uV/K

# Exact Lorentz number
        L1 = (mobility.rfac + 1.0) * (mobility.rfac + 3.0) * Fr * Fr2
        L2 = pow((mobility.rfac + 2.0) * Fr1, 2)
        L3 = pow((mobility.rfac + 1.0) * Fr, 2)
        L  = kBe2 * (L1 - L2) / L3

        kappa = L * sigma / 0.01 * T
#        L2 = kappa / (sigma * 100.0) / T

#        print("kappa={}/{} W/m/K".format(kappa, kappa2))
        if validate_error_str is not None and validate_error_str != '' and abs(kappa) > 1.0e-4:
            Lk1 = (mobility.rfac + 3.0) * Fr2
            Lk2 = (mobility.rfac + 2.0)**2 * Fr1 * Fr1
            Lk3 = (mobility.rfac + 1.0) * Fr
            kappa2 = kkappa * pow(kBT, mobility.rfac + 2.0) * (Lk1 - Lk2 / Lk3)           # W/m/K
            validate_error(kappa, kappa2, 1.0e-8, "{} for kappa: ".format(validate_error_str))

        kappa_tot = kappa + klatt
        PF = pow(S * 1e-6, 2) * (sigma / 0.01)                                  # W/m/K2
        ZT = PF * T / kappa_tot
        PF *= 1.0e6 / 0.01                                                      # uW/cm/K2

        sigma *= charge*charge
        mu    *= abs(charge)
        S     /= charge
        
        inf = {
              "sigma": sigma, "n": n, "mu": mu, "<tau>": tau_avg, "S": S, "kappa,e": kappa, "kappa,tot": kappa_tot
            , "L": L, "PF": PF, "ZT": ZT
            }
            
        return sigma, n, mu, tau_avg, S, kappa, kappa_tot, L, PF, ZT, inf


    def cal_LorentzNumber_from_S_approximate(self, S):
        return (1.5 + exp(-S / 116)) * 1.0e-8

    def cal_S_nondegenerated_from_Ne(self, n, rfac, charge = 1.0): # V/K
        return kB / e * (log(self.NC / n) + rfac + 2.0)  / charge

    def cal_S_degenerated_from_Ne(self, T, n, rfac, charge = 1.0):  # V/K
        return kS3 * self.meeff * me * T * pow(n * 1.0e6, -2.0/3.0) * (rfac + 1.0) / charge

# Hall-related properties for single carrier model
    def cal_Hall_properteis(self, T, EF, mobility, validate_error_str = None, charge = None):
        if charge is None:
            charge = mobility.charge

        x  = EF * e / kB / T

        mobility.set_scattering_parameters()
#        self.NV  = meff2NC_FEA(self.mheff, T)
        self.NC  = meff2NC_FEA(self.meeff, T)
#        self.DV0 = meff2DC0_FEA(self.mheff, T)
        self.DC0 = meff2DC0_FEA(self.meeff, T)

        ksigma = ksigma0 * mobility.l0norm * self.meeff
        kmu    = kmu0 * mobility.l0norm / sqrt(self.meeff) 
        ktau   = ktau0 * mobility.l0norm * (mobility.rfac + 1.0) * sqrt(self.meeff) 

        kBT = kB * T

        F05 = FermiIntegral(x, 0.5)
        Fr  = FermiIntegral(x, mobility.rfac)
        Fr1 = FermiIntegral(x, mobility.rfac + 1.0)
        Fr2 = FermiIntegral(x, mobility.rfac + 2.0)
# Carrier density, conductivity, mobility
        n     = self.NC * kpi05 * F05                                                # cm^-3
        sigma = ksigma * pow(kBT, mobility.rfac + 1.0) * (mobility.rfac + 1.0) * Fr / 100.0   # S/cm
        mu    = sigma / e / n                                                   # cm2/Vs

# Relaxation time average
        tau_avg = mu * 1.0e-4 * self.meeff * me / e * 1.0e15                          # fs
        if validate_error_str is not None and validate_error_str != '':
            tau_avg2 = ktau * pow(kBT, mobility.rfac - 0.5) * Fr / F05 * 1.0e15
            validate_error(tau_avg, tau_avg2, 1.0e-10, "{} for <tau>: ".format(validate_error_str))

        F2rhalf = FermiIntegral(x, 2.0 * mobility.rfac - 0.5)
        FH      = 1.5 * (2.0 * mobility.rfac + 0.5) / (mobility.rfac + 1.0)**2 * F05 * F2rhalf / Fr / Fr
        RH0     = 1.0 / e / n / charge
        RH      = FH * RH0
        n_Hall  = n / FH
        mu_Hall = FH * mu

#        print("mu={}/{} cm2/Vs".format(mu, mu2))
        if validate_error_str is not None and validate_error_str != '':
            mu2 = kmu * pow(kBT, mobility.rfac - 0.5) * (mobility.rfac + 1.0) * Fr / F05 *1.0e4 # cm2/Vs
            validate_error(mu, mu2, 1.0e-10, "{} for mu: ".format(validate_error_str))

        sigma *= charge*charge
        mu    *= abs(charge)

        inf = {
              "sigma": sigma, "n": n, "mu": mu, "<tau>": tau_avg,
              "FH": FH, "RH": RH, "RH0": RH0, "nHall": n_Hall, "muHall": mu_Hall
            }
            
        return inf

    def cal_EF_lists(self, T, xEF, print_level = 0):
        nEF    = len(xEF)
        yNe    = []
        yNh    = []
        yNDp   = []
        yNAm   = []
        yRH    = []
        yNs    = []
        yNsabs = []

        EFprev = self.EF0
        kBTe = kB * T / e
        dEinteg = self.nrange * kBTe
        E0h = self.EV - dEinteg
        E1e = self.EC + dEinteg
#    print("dEinteg=", dEinteg, T0)
#    exit()

        if print_level >= 1:
            print("")
            print("at {} K".format(T))
            print ("%5s\t%8s\t%14s\t%14s\t%14s\t%14s\t%14s" % ("EF(eV)", "Ne", "Nh", "ND+", "NA-", "RH", "Ns"))
        for iEF in range(nEF):
            EF = xEF[iEF]
            Nh1, Ne1, NAm1, NDp1, RH1, Ns1 = self.cal_carrier_densities(T, EF, E0h, self.EV, self.EC, E1e, xEF[1] - xEF[0])
#            inf = self.cal_Hall_properteis2(T, EF, mobility, validate_error_str = None, charge = None)
#        inf = {
#              "sigma": sigma, "n": n, "mu": mu, "<tau>": tau_avg,
#              "FH": FH, "RH": RH, "RH0": RH0, "nHall": n_Hall, "muHall": mu_Hall
#            }
#            Nh1, Ne1, NAm1, NDp1, RH1, Ns1 = inf["nh"], inf["ne"], inf["NAm"], inf["NDp"], inf["RH"], inf["ns"]

            yNe.append(Ne1)
            yNh.append(Nh1)
            yNDp.append(NDp1)
            yNAm.append(NAm1)
            yRH.append(RH1)
            yNs.append(Ns1)
            yNsabs.append(abs(Ns1))
            if print_level >= 1:
                print ("%10.4f\t%12.6g\t%12.6g\t%12.6g\t%12.6g\t%12.6g\t%12.6g" % (EF, Ne1, Nh1, NDp1, NAm1, RH1, Ns1))

        return yNh, yNe, yNAm, yNDp, yRH, yNs, yNsabs

    def cal_T_lists(self, xT, print_level = 0):
        xTinv = []
        yEF   = []
        yNe    = []
        yNh    = []
        yNDp   = []
        yNAm   = []
        yRH    = []
        yNs    = []
        yNsabs = []
        ydQ    = []
        EFprev = self.EF0
#初期範囲
        EFmin = min([self.EV - 0.1, self.EF0 - self.dE_bisec])
        EFmax = max([self.EC + 0.1, self.EF0 + self.dE_bisec])
#    dEFbisec = 0.1
        if print_level >= 1:
            print("")
        for iT in range(len(xT)):
            T = xT[iT]

            if print_level >= 1:
                print("iT {:03d}: {} K".format(iT, T))

            kBTe = kB * T / e
            dEinteg = self.nrange * kBTe

            E0h = self.EV - dEinteg
            E1e = self.EC + dEinteg

#            if E0h < self.Emin or self.Emax < E1e:
#                if print_level >= 1:
#                    print("*** Integration range error: [{}, {}] exceeds DOS E range [{}, {}]".format(E0h, E1e, self.Emin, self.Emax))
#                exit()

            self.NC  = self.meff2NC_FEA(self.meeff, T)
            self.NV  = self.meff2NC_FEA(self.mheff, T)
            self.DC0 = self.meff2DC0_FEA(self.meeff, T)
            self.DV0 = self.meff2DC0_FEA(self.mheff, T)

#        Ne0K = self.Ne(0.0, self.EF0, E0h, self.EF0, self.Estep)
            EF, dQ = self.FindEF_semi(T, EFmin, EFmax, E0h, self.EV, self.EC, E1e, self.Estep)
            Nhh, Neh, NAmh, NDph, RH, Ns = self.cal_carrier_densities(T, EF, E0h, self.EV, self.EC, E1e, self.Estep)

            xTinv.append(1000.0 / T)
            yEF.append(EF)
            yNe.append(Neh)
            yNh.append(Nhh)
            yNDp.append(NDph)
            yNAm.append(NAmh)
            yRH.append(RH)
            yNs.append(Ns)
            yNsabs.append(abs(Ns))
            ydQ.append(dQ)
            if print_level >= 1:
                print("  ***(T,EF,Ne,Nh,NDp,NAm,RH)=({:6.4f},{:6.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                            .format(T, EF, Neh, Nhh, NDph, NAmh, RH, Ns))

            EFmin = min([self.EV, EF - self.dE_bisec])
            EFmax = max([self.EC, EF + self.dE_bisec])
    
        return xTinv, yEF, yNe, yNh, yNAm, yNDp, yRH, yNs, yNsabs, ydQ


def main():
    print("")
    print("This is a library, not runnable")
    print("")


if __name__ == "__main__":
    main()
