import numpy as np
from numpy import exp
import pandas as pd
from scipy.optimize import minimize


#nelder-mead    Downhill simplex
#powell         Modified Powell
#cg             conjugate gradient (Polak-Ribiere method)
#bfgs           BFGS法
#newton-cg      Newton-CG
#trust-ncg      信頼領域 Newton-CG 法
#dogleg         信頼領域 dog-leg 法
#L-BFGS-B’ (see here)
#TNC’ (see here)
#COBYLA’ (see here)
#SLSQP’ (see here)
#trust-constr’(see here)
#dogleg’ (see here)
#trust-exact’ (see here)
#trust-krylov’ (see here)

method = "bfgs"


#==========================================
# Source parameters to be fitted
#==========================================
infile = 'peak.xlsx'
nx = None

# optimization parameters
x0s = [1.3,  0.6, 0.1]

# 数値微分する際の変数の微小変位
h = 0.01

maxiter = 100
tol = 1.0e-5



#==========================================
# functions
#==========================================
def cal_y(x, params):
    I0 = params[0]
    x0 = params[1]
    w  = params[2]
    a = 0.832554611 / w
    X = a * (x - x0)
    return I0 * exp(-X * X)

def ycal_list(xd, params):
    y = []
    for i in range(len(xd)):
        y.append(cal_y(xd[i], params))
    return y

def minimize_func(xk, xd, yd):
    nd = len(xd)
    MSE = 0.0
    for i in range(nd):
        yc = cal_y(xd[i], xk)
        d  = yd[i] - yc
        MSE += d * d

    return MSE / nd

# １次微分を定義するとcgやbfgsなどの勾配法を使える
def diff1(x, xd, yd):
    diff = x.copy()
    nvar = len(x)

    for i in range(nvar):
        xx = x.copy()
        xx[i] = x[i] - h
        ym = minimize_func(xx, xd, yd)
        
        xx[i] = x[i] + h
        yp = minimize_func(xx, xd, yd)
        diff[i] = (yp - ym) / 2.0 / h

    return diff


# callbackを使うと、最適化過程をモニターできる
# 引数は変数のリストだけが渡される
# 反復回数などはglobal変数で保持する
iter = 0
def callback(xk, xd, yd):
    global iter

    fmin = minimize_func(xk, xd, yd)
    print(f"callback {iter}: xk={xk}  func={fmin}")
    iter += 1

#==========================================
# Main routine
#==========================================
def main():
    print("")
    print("Simplest peak fitting program by simplex")
    print(f"method        : {method}")
    print(f"input file    : {infile}")
    print(f"initial values: {x0s}")
    print(f"tol           : {tol}")
    print(f"maxiter       : {maxiter}")

    print("")
    print(f"Read [{infile}]")
    df = pd.read_excel(infile, engine = 'openpyxl')
    labels = df.columns.to_list()
    xd = df[labels[0]].to_numpy()
    yd = df[labels[1]].to_numpy()
    nx = len(xd)
    xrange = [min(xd), max(xd)]

    res = minimize(lambda xk: minimize_func(xk, xd, yd), x0s, 
                    method = method, 
                    jac = lambda xk: diff1(xk, xd, yd), 
                    callback = lambda xk: callback(xk, xd, yd),
                    tol = tol, options = {'maxiter':maxiter, "disp":True})
#    print("")
#    print(res)
    print("")
    if res.success:
        print(f"Function MSE takes the minimum")
        print(f"   at MSE={res.fun}")
        print(f"   with x={res.x}")
        print(f" iteration: {res.nit}")
    else:
        print(f"Function did not converge")
        print(res)


if __name__ == "__main__":
    main()
